(*
From: Gregory Morrisett <jgm@CS.Cornell.EDU>
To: Matthew Fluet <fluet@CS.Cornell.EDU>
Subject: ML in ML
*)
infix ++;

signature LAMBDA =
  sig
    (* i.e., G |- exp : a *)
    type ('G,'a) exp

    type Top

    (* 'a <= 'b *)
    type ('a,'b) Sub
    val top : ('a,Top) Sub
    val intreal : (int,real) Sub
    val refl : ('a,'a) Sub
    val tran : ('a,'b) Sub * ('b,'c) Sub -> ('a,'c) Sub
    val arrowcon : ('c,'a) Sub * ('b,'d) Sub -> ('a->'b,'c->'d) Sub
    val paircon : ('a,'b) Sub * ('c,'d) Sub -> ('a*'b,'c*'d) Sub

    (* G |- i : int *)
    val int : int -> ('G,int) exp
    (* G |- f : real *)
    val real : real -> ('G,real) exp
    (* G |- e : real 
       ---------------------
       G |- floor e : int
     *)
    val floor : ('G,real) exp -> ('G,int) exp

    (* G |- e1 : real    G |- e2 : real
       ----------------------------------
       G |- e1 + e2 : real
     *)
    val ++ : ('G,real) exp * ('G,real) exp -> ('G,real) exp

    (* G,Z:a |- Z : a *)
    val Z : ('G*'a,'a) exp

    (* G |- exp : a  
       ------------------------
       G,Y:b |- W exp : a
     *)
    val W : ('G,'a) exp -> ('G*'b,'a) exp

    (* G,X:a,Y:b |- exp : c
       -------------------------
       G,Y:b,X:b |- S exp : c
     *)
    val S : (('G*'a)*'b,'c) exp -> (('G*'b)*'a,'c) exp

    (* G |- exp1 : a->b     G |- exp2 : a
       ----------------------------------
       G |- exp1(exp2) : b
     *)
    val app : ('G,'a->'b) exp * ('G,'a) exp -> ('G,'b) exp

    (* G,a |- exp : b
       -----------------
       G |- \.exp : a->b
     *)
    val lambda : ('G*'a,'b) exp -> ('G,'a->'b) exp

    (* G |- exp1 : 'a   G |- exp2 : 'b
       -------------------------------
       G |- (exp1,exp2) : 'a*'b
     *)
    val pair : ('G,'a) exp * ('G,'b) exp -> ('G,'a*'b) exp

    (* G |- exp : 'a*'b
       -----------------
       G |- pi1 : exp : 'a
       G |- pi2 : exp : 'b
     *)
    val pi1 : ('G,'a*'b) exp -> ('G,'a) exp
    val pi2 : ('G,'a*'b) exp -> ('G,'b) exp

    (* G |- exp : 'a    'a <= 'b
       -------------------------
       G |- exp : 'b
     *)
    val subsume : ('G,'a) exp * ('a,'b) Sub -> ('G,'b) exp

    val eval : (unit,int) exp -> int
  end

structure Lambda :> LAMBDA =
  struct
    type Top = unit
    datatype Exp =
      Int_e of int
    | Real_e of real
    | Floor_e of Exp
    | Plus_e of Exp * Exp
    | Z_e
    | W_e of Exp
    | S_e of Exp
    | App_e of Exp * Exp
    | Lambda_e of Exp
    | Pair_e of Exp * Exp
    | Pi1_e of Exp
    | Pi2_e of Exp

    type ('G,'a) exp = Exp
    type ('a,'b) Sub = unit
    val refl = ()
    val tran = fn _ => ()
    val arrowcon = fn _ => ()
    val paircon = fn _ => ()
    val top = ()
    val intreal = ()

    val int = Int_e;
    val real = Real_e;
    val floor = Floor_e;
    val (op ++) = Plus_e;
    val Z = Z_e;
    val W = W_e;
    val S = S_e;
    val app = App_e;
    val lambda = Lambda_e;
    val pair = Pair_e;
    val pi1 = Pi1_e;
    val pi2 = Pi2_e;
    val subsume = fn (e,_) => e

    datatype Val = 
      Int_v of int | Real_v of real | Fn_v of Val -> Val 
    | Pair_v of Val * Val

    exception Impossible
    fun deReal (Real_v f) = f
      | deReal (Int_v i)  = Real.fromInt i
      | deReal _ = raise Impossible
    fun deFn (Fn_v f) = f
      | deFn _ = raise Impossible
    fun dePair (Pair_v p) = p
      | dePair _ = raise Impossible

    type Env = Val list

    fun eval' (env: Val list) : Exp -> Val = 
      fn Int_e i => Int_v i
       | Real_e f => Real_v f
       | Floor_e e => 
          (case eval' env e of
             Int_v i => Int_v i
           | Real_v f => Int_v(Real.floor f)
           | _ => raise Impossible)
       | Plus_e(e1,e2) => Real_v(deReal(eval' env e1) + deReal(eval' env e2))
       | Z_e => hd env
       | W_e e => eval' (tl env) e
       | S_e e => eval' ((hd(tl env))::(hd env)::(tl(tl env))) e
       | App_e(e1,e2) => (deFn(eval' env e1))(eval' env e2)
       | Lambda_e e => Fn_v (fn v => eval' (v::env) e)
       | Pair_e(e1,e2) => Pair_v(eval' env e1, eval' env e2) 
       | Pi1_e e => #1(dePair(eval' env e))
       | Pi2_e e => #2(dePair(eval' env e))

    fun eval e = 
      case eval' [] e of
        Int_v i => i
      | _ => raise Impossible
  end

functor TestLambda(Lambda : LAMBDA) = 
struct
  local
    open Lambda;
    val x = W Z
    val y = Z
    val f = lambda(lambda (x ++ subsume(y,intreal)))
    val arg1 = real 3.1415
    val arg2 = int 2
  in
    val eval = eval
    val exp : (unit,int) exp = floor(app(app(f,arg1),arg2))
  end
end

structure Test = TestLambda(Lambda);
